#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "callbacks.h"
#include "gnusudoku_gui.h"
#include "support.h"

#include "sudoku.h"

guint contextid = 0;
int emptycelcount = -1;
char sudokufile[256]={0};
int symmetric = 1;
int popuptips = 1;
static volatile int pulsing = 0;
extern int keepgoing;
void update_statusbar1(const char *buf);

static int pulsar(void)
{
	gtk_progress_bar_pulse(GTK_PROGRESS_BAR(progressbar1));
	if( gtk_main_iteration_do(FALSE) )
		return -1;
	return pulsing;
}

void update_label1(const char *txt)
{
	char buf[80];
	sprintf(buf,"<b>%s</b>",basename(txt));
	gtk_label_set_markup(GTK_LABEL(label1),buf);
}

void
open_file(const char *name)
{
	FILE *fp;
	char line[80],*validchars;
	int tx,ty,cx,cy,i,j,k,sudokuwidth;
	GdkEventKey event;
	fp = fopen(name,"r");
	if( !fp ) return;
try_again:
	switch(sudokusize) {
		case 2:
			validchars = "1234";
			break;
		case 3:
		default:
			validchars = "123456789";
			break;
		case 4:
			validchars = "0123456789ABCDEF";
			break;
		case 5:
			validchars = "ABCDEFGHIJKLMNOPQRSTUVWXY";
			break;
	}
	on_clear_activate(0, 0);
	sudokuwidth = sudokusize * sudokusize;
	i = 0;
	while( i<sudokuwidth ) {
		if( !fgets(line,80,fp) )
			break;
		if( line[0] == '+' )
			continue;
		if( i==0 && line[0] == '#' ) {
			if( isdigit(line[1]) ) {
				j = line[1] - '0';
				if( j!=sudokusize && j>1 && j<6 ) {
					keepgoing = 1;
					gtk_widget_hide (windowGnuSudoku);
					gtk_widget_destroy(windowGnuSudoku);
					sudokusize = j;
					windowGnuSudoku = create_windowGnuSudoku ();
					gtk_widget_show (windowGnuSudoku);
					goto try_again;
				}
			}
			continue;
		}
		cy = i % sudokusize;
		ty = i / sudokusize;
		j = 0;
		k = 0;
		while( k<sudokuwidth ) {
			if( line[j] != '|' ) {
				if( strchr(validchars,toupper(line[j])) && line[j] ) {
					cx = k % sudokusize;
					tx = k / sudokusize;
					event.keyval = line[j];
					gtk_widget_grab_focus(button[ty][tx][cy][cx]);
					event.type = GDK_KEY_PRESS;
					event.window = gtk_widget_get_root_window(windowGnuSudoku);
					event.send_event = TRUE;
					on_button_key_press_event(button[ty][tx][cy][cx],&event,0);
				}
				k++;
			}
			j++;
		}
		i++;
	}
	fclose(fp);
	strcpy(sudokufile,name);
	update_label1(sudokufile);
}

static void save_separator(FILE *fp)
{
	int i,j;
	for( i=0; i<sudokusize; i++ ) {
		fputc('+',fp);
		for( j=0; j<sudokusize; j++ )
			fputc('-',fp);
	}
	fputc('+',fp);
	fputc('\n',fp);
}

void
save_file(const char *name)
{
	FILE *fp;
	const char *p;
	int tx,ty,cx,cy;
	GdkEventKey event;
	fp = fopen(name,"w");
	if( !fp ) return;
	strcpy(sudokufile,name);
	update_label1(sudokufile);
	fprintf(fp,"#%dx%d Sudoku\n",sudokusize,sudokusize);
	for( ty=0; ty<sudokusize; ty++ ) {
		save_separator(fp);
		for( cy=0; cy<sudokusize; cy++ ) {
			for( tx=0; tx<sudokusize; tx++ ) {
				fputc('|',fp);
				for( cx=0; cx<sudokusize; cx++ ) {
					p = gtk_button_get_label(GTK_BUTTON(button[ty][tx][cy][cx]));
					if( *p == ' ' )
						fputc('.',fp);
					else
						fputc(*p,fp);
				}
			}
			fputc('|',fp);
			fputc('\n',fp);
		}
	}
	save_separator(fp);
	fclose(fp);
}

void
on_new1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	sudokufile[0] = 0;
	on_clear_activate(menuitem, user_data);
	update_label1("noname");
}


void
on_open1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkWidget *dialog;
	dialog = create_filechooserdialog();
	if( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)
  {
    char *filename;

    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    open_file(filename);
    g_free (filename);
		pulsing = 0;
  }
	gtk_widget_destroy(dialog);
}


void
on_save1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	if( pulsing )
		return;
	if( sudokufile[0] )
		save_file(sudokufile);
	else
		on_save_as1_activate(menuitem, user_data);
}


void
on_save_as1_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkWidget *dialog;
	if( pulsing )
		return;
	dialog = create_filesaveasdialog();
	if( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)
  {
    char *filename;

    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    save_file(filename);
    g_free (filename);
  }
	gtk_widget_destroy(dialog);
}


void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}

static char *sudokustring(int minlen)
{
	char *buf,*p;
	int cx,cy,tx,ty,z;
	z = sudokusize*sudokusize*sudokusize*sudokusize+sudokusize*sudokusize+1;
	if( z<minlen )
		z = minlen;
	buf = malloc(z);
	if( !buf )
		return 0;
	p = buf;
	for( ty=0; ty<sudokusize; ty++ )
		for( cy=0; cy<sudokusize; cy++ ) {
			for( tx=0; tx<sudokusize; tx++ )
				for( cx=0; cx<sudokusize; cx++ ) {
					z = *gtk_button_get_label(GTK_BUTTON(button[ty][tx][cy][cx]));
					if( ispunct(z) )
						*p++ = ' ';
					else
						*p++ = toupper(z);
				}
			*p++ = '\n';
		}
	*p = '\0';
	return buf;
}

void
on_hint_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	char *buf, *p;
	int cx,cy,tx,ty,z;
	if( pulsing )
		return;
	buf = sudokustring(60);
	if( !buf )
		return;
	init_sudoku(sudokusize, buf);
	z = travelsudoku();
	if( z ) {
		ty = Y_DECODED(z);
		tx = X_DECODED(z);
		cy = ty % sudokusize;
		cx = tx % sudokusize;
		gtk_widget_grab_focus(button[ty/sudokusize][tx/sudokusize][cy][cx]);
		gtk_label_set_text(GTK_LABEL(labelExplain),explanation(z));
		if( sudokuvalid )
			sprintf(buf,_("Hint:%c on selected cel row %d col %d"),V_DECODED(z),ty+1,tx+1);
		else
			sprintf(buf,_("Unsolvable:%c on cel row %d col %d"),V_DECODED(z),ty+1,tx+1);
	}
	else {
		if( sudokuvalid )
			sprintf(buf,_("%dx%d sudoku - no hint available"),sudokusize,sudokusize);
		else
			sprintf(buf,_("%dx%d sudoku - is unsolvable"),sudokusize,sudokusize);
	}
	update_statusbar1(buf);
	free(buf);
}

void
update_progressbar1(void)
{
	char buf[6];
	int cels = sudokusize * sudokusize * sudokusize * sudokusize;
	gdouble fraction = ((gdouble)(cels - emptycelcount)) / (gdouble)cels;
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progressbar1), fraction);
	sprintf(buf,"%d%%",(cels - emptycelcount)*100/cels);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar1), buf);
}

void
update_statusbar1(const char *buf)
{
	if( contextid == 0 ) {
		contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar1),"sudoku");
		gtk_statusbar_push(GTK_STATUSBAR(statusbar1), contextid, "");
	}
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar1), contextid);
	gtk_statusbar_push(GTK_STATUSBAR(statusbar1), contextid, buf);
}

void
on_clear_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	int tx,ty,cx,cy;
	char buf[60];
	switch( sudokusize ) {
		case 5:
			strcpy(buf,_("A B C D E F G H I J K L M N O P Q R S T U V W X Y"));
			break;
		case 4:
			strcpy(buf,_("0 1 2 3 4 5 6 7 8 9 A B C D E F"));
			break;
		case 3:
		default:
			strcpy(buf,_("1 2 3 4 5 6 7 8 9"));
			break;
		case 2:
			strcpy(buf,_("1 2 3 4"));
			break;
	}
	for( tx=0; tx<sudokusize; tx++ )
		for( ty=0; ty<sudokusize; ty++ )
			for( cx=0; cx<sudokusize; cx++ )
				for( cy=0; cy<sudokusize; cy++ ) {
					gtk_button_set_label(GTK_BUTTON(button[ty][tx][cy][cx]),EMPTY_CEL);
					gtk_tooltips_set_tip (celtooltips, button[ty][tx][cy][cx], buf, NULL);
				}
	sprintf(buf,_("%dx%d sudoku cleared"),sudokusize,sudokusize);
	update_statusbar1(buf);
	emptycelcount = sudokusize * sudokusize * sudokusize * sudokusize;
	update_progressbar1();
	gtk_label_set_text(GTK_LABEL(labelExplain),"");
	pulsing = 0;
}

void
on_solve_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	char *buf, *p;
	int cx,cy,tx,ty,z,grade;
	GdkEventKey event;
	if( pulsing )
		return;
	buf = sudokustring(60);
	if( !buf )
		return;
	init_sudoku(sudokusize, buf);
	z=1;
	grade = 0;
	while( z && unsolved() && sudokuvalid ) {
		z = travelsudoku();
		if( z ) {
			grade += G_DECODED(z);
			ty = Y_DECODED(z);
			tx = X_DECODED(z);
			cy = ty % sudokusize;
			ty = ty / sudokusize;
			cx = tx % sudokusize;
			tx = tx / sudokusize;
			gtk_widget_grab_focus(button[ty][tx][cy][cx]);
			event.keyval = V_DECODED(z);
			event.type = GDK_KEY_PRESS;
			event.window = gtk_widget_get_root_window(windowGnuSudoku);
			event.send_event = TRUE;
			on_button_key_press_event(button[ty][tx][cy][cx],&event,0);
			gtk_main_iteration_do(FALSE);
		}
		else {
			if( sudokuvalid )
				sprintf(buf,_("%dx%d sudoku - no hint available"),sudokusize,sudokusize,V_DECODED(z),cx,cy,tx,ty);
			else
				sprintf(buf,_("%dx%d sudoku - is unsolvable"),sudokusize,sudokusize,V_DECODED(z),cx,cy,tx,ty);
			grade = 0;
			update_statusbar1(buf);
		}
	}
	if( grade ) {
		sprintf(buf,"This was a grade %d sudoku",grade);
		gtk_label_set_text(GTK_LABEL(labelExplain),buf);
	}
	else
		gtk_label_set_text(GTK_LABEL(labelExplain),"");
	free(buf);
}


void
on_setup_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkWidget *d;
	int rib;
	d = create_setupdialog();
	if( gtk_dialog_run(GTK_DIALOG(d)) == GTK_RESPONSE_OK ) {
		popuptips = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (checkbuttonTips));
		symmetric = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (checkbuttonSymmetry));
		rib = 3;
		if( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radiobutton2x2)) ) rib = 2;
		if( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radiobutton3x3)) ) rib = 3;
		if( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radiobutton4x4)) ) rib = 4;
		if( gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radiobutton5x5)) ) rib = 5;
		if( rib != sudokusize ) {
			keepgoing = 1;
			gtk_widget_hide (windowGnuSudoku);
			gtk_widget_destroy(windowGnuSudoku);
			sudokusize = rib;
			windowGnuSudoku = create_windowGnuSudoku ();
			gtk_widget_show (windowGnuSudoku);
		}
		if( popuptips )
			gtk_tooltips_enable(celtooltips);
		else
			gtk_tooltips_disable(celtooltips);
		strcpy(helpbrowser, gtk_entry_get_text(GTK_ENTRY(entryBrowser)));
	}
	gtk_widget_destroy(d);
}


void
on_about1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	GtkWidget *d;
	d = create_aboutdialog();
	gtk_dialog_run(GTK_DIALOG(d));
	gtk_widget_destroy(d);
}


void
on_buttonQuit_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_button_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
	GdkEventKey event;
	if( !popuptips )
		return;
	event.keyval = '=';
	event.type = GDK_KEY_PRESS;
	event.window = gtk_widget_get_root_window(windowGnuSudoku);
	event.send_event = TRUE;
	on_button_key_press_event(GTK_WIDGET(button),&event,0);
}


void
on_button_notify                       (GObject         *gobject,
                                        GParamSpec      *arg1,
                                        gpointer         user_data)
{

}

void
add_to_tooltip(GtkWidget *widget, char *p)
{
	GtkTooltipsData *tip;
	char tiptxt[60];
	tip = gtk_tooltips_data_get(widget);
	sprintf(tiptxt,"%s %s",p,tip->tip_text);
	gtk_tooltips_set_tip (celtooltips, widget, tiptxt, NULL);
}

void
delete_from_tooltip(GtkWidget *widget, char *t)
{
	char *p;
	GtkTooltipsData *tip;
	char tiptxt[60];
	tip = gtk_tooltips_data_get(widget);
	strcpy(tiptxt, tip->tip_text);
	if( (p=strchr(tiptxt,*t)) ) {
		if( p[1] )
			strcpy(p,p+2);
		else {
			if( p > tiptxt )
				p--;
			*p = '\0';
		}
		gtk_tooltips_set_tip (celtooltips, widget, tiptxt, NULL);
	}
}

gboolean in_row(int ty, int cy, char *p)
{
	int cx,tx;
	for( tx=0; tx<sudokusize; tx++ )
		for( cx=0; cx<sudokusize; cx++ )
			if( strcmp(gtk_button_get_label(GTK_BUTTON(button[ty][tx][cy][cx])),p) == 0 )
				return TRUE;
	return FALSE;
}

gboolean in_column(int tx, int cx, char *p)
{
	int ty,cy;
	for( cy=0; cy<sudokusize; cy++ )
		for( ty=0; ty<sudokusize; ty++ )
			if( strcmp(gtk_button_get_label(GTK_BUTTON(button[ty][tx][cy][cx])),p) == 0 )
				return TRUE;
	return FALSE;
}

gboolean in_square(int ty, int tx, char *p)
{
	int cx,cy;
	for( cy=0; cy<sudokusize; cy++ )
		for( cx=0; cx<sudokusize; cx++ )
			if( strcmp(gtk_button_get_label(GTK_BUTTON(button[ty][tx][cy][cx])),p) == 0 )
				return TRUE;
	return FALSE;
}

gboolean
on_button_key_press_event              (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
	char buf[60];
	const char *p;
	int tx,ty,cx,cy,i,j;
	if( pulsing )
		return;
	for( p = gtk_widget_get_name(widget); !isdigit(*p); p++ )
		;
	ty = p[0]-'0';
	tx = p[1]-'0';
	cy = p[2]-'0';
	cx = p[3]-'0';
	if( emptycelcount < 0 ) {
		emptycelcount = sudokusize * sudokusize * sudokusize * sudokusize;
	}
	strcpy(buf,gtk_button_get_label(GTK_BUTTON(widget)));
	if( popuptips && (event->keyval == '=') ) {
		p=gtk_tooltips_data_get(widget)->tip_text;
		if( p[1]=='\0' )
			event->keyval = p[0];
	}
	if( (event->keyval == ' ') || (p=strchr(gtk_tooltips_data_get(widget)->tip_text,toupper(event->keyval))) ) {
		gtk_button_set_label(GTK_BUTTON(widget),EMPTY_CEL);
		if( strcmp(buf,EMPTY_CEL) ) {
			emptycelcount++;
			add_to_tooltip(widget,buf);
			for( i=0; i<sudokusize; i++ )
				for( j=0; j<sudokusize; j++ ) {
					/* square */
					if( j != cx || i != cy )
						if( !(in_column(tx,j,buf) || in_row(ty,i,buf)) )
							add_to_tooltip(button[ty][tx][i][j],buf);
					/* row */
					if( j != tx && !in_square(ty,j,buf) )
						add_to_tooltip(button[ty][j][cy][i],buf);
					/* column */
					if( i != ty && !in_square(i,tx,buf) )
						add_to_tooltip(button[i][tx][j][cx],buf);
				}
		}
	}
	if( event->keyval != ' ' ) {
		if( p ) {
			emptycelcount--;
			sprintf(buf,"%c",toupper(event->keyval));
			gtk_button_set_label(GTK_BUTTON(widget),buf);
			delete_from_tooltip(widget,buf);
			for( i=0; i<sudokusize; i++ )
				for( j=0; j<sudokusize; j++ ) {
					/* square */
					if( j != cx || i != cy )
						delete_from_tooltip(button[ty][tx][i][j],buf);
					/* row */
					if( j != tx )
						delete_from_tooltip(button[ty][j][cy][i],buf);
					/* column */
					if( i != ty )
						delete_from_tooltip(button[i][tx][j][cx],buf);
				}
		}
	}
	if( emptycelcount > 0 )
		sprintf(buf,_("%dx%d sudoku - %d cels left"),sudokusize,sudokusize,emptycelcount);
	else
		sprintf(buf,_("%dx%d sudoku solved"),sudokusize,sudokusize,emptycelcount);
	update_statusbar1(buf);
	update_progressbar1();
	gtk_label_set_text(GTK_LABEL(labelExplain),"");
  return FALSE;
}

gboolean
on_button_key_release_event            (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
  return FALSE;
}


gboolean
on_button_enter_notify_event           (GtkWidget       *widget,
                                        GdkEventCrossing *event,
                                        gpointer         user_data)
{
	gtk_widget_grab_focus(widget);
  return FALSE;
}

/* choice -- returns cyclic via cycle from the 3! possible combinations of a,b and c */
static int choice(int cycle,int a,int b,int c)
{
	switch(cycle % (1*2*3)) {
		case 0: return a;
		case 1: return b;
		case 2: return a;
		case 3: return c;
		case 4: return c;
		case 5: return b;
		default:
			break;
	}
	return a;
}

void
on_generate1_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	int e,i,j,w,x,y,z,k,grade,maximumcels;
	char c[32];	/* watch it! candidates are initialized with a strcpy so preserve space for the terminating zero */
	char *buf,*solvedbuf;
	GdkEventKey event;
	if( pulsing ) {	/* another press while generating, stop generating */
		pulsing = 0;
		return;
	}
	buf = sudokustring(0);
	if( !buf )
		return;
	gtk_progress_bar_set_pulse_step(GTK_PROGRESS_BAR(progressbar1),0.01);
	pulsing = 1;
	init_sudoku(sudokusize, buf);
	z=1;
	w = sudokusize * sudokusize;
	k = symmetric;
	e=w*w;
	for( i=0; buf[i]; i++ )
		if( buf[i] == ' ' )
			--e;
	if( e )
		solvedbuf = buf;
	else {
		gtk_label_set_text(GTK_LABEL(labelExplain),_("calculating sudoku..."));
		generatesolved(pulsar);
		solvedbuf = sudokulines();
		if( !solvedbuf )
			solvedbuf = buf;
		init_sudoku(sudokusize, buf);
	}
	for( ;pulsing; ) {
		sprintf(c,_("cels generated: %d"),e);
		gtk_label_set_text(GTK_LABEL(labelExplain),c);
		if( pulsar() < 0 )
			goto panic;
		grade = 0;
		while( z && unsolved() && sudokuvalid && pulsing ) {
			z = travelsudoku();
			if( pulsar() < 0 )
				goto panic;
			grade += G_DECODED(z);
		}
		if( pulsing ) {
			if( sudokuvalid ) {
				if( z ) /* a solution can be found, so this is a good sudoku */
					break;
				/* no hint available , need to generate another number */
				if( k>0 ) {	/* go for symmetry */
					for( y=0; y<w; y++ ) {
						for( x=0; x<w; x++ ) {
							if( buf[y*(w+1)+x] != ' ' ) {
								k = picknumber(6);
								i = choice(k,w-y-1,w-y-1,y);
								j = choice(k,x,w-x-1,w-x-1);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								i = choice(k,w-y-1,y,w-y-1);
								j = choice(k,w-x-1,w-x-1,x);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								i = choice(k,y,w-y-1,w-y-1);
								j = choice(k,w-x-1,x,w-x-1);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								j = choice(k,w-y-1,w-y-1,y);
								i = choice(k,x,w-x-1,w-x-1);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								j = choice(k,w-y-1,y,w-y-1);
								i = choice(k,w-x-1,w-x-1,x);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								j = choice(k,y,w-y-1,w-y-1);
								i = choice(k,w-x-1,x,w-x-1);
								if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) ) goto choose;
								if( pulsar() < 0 )
									goto panic;
							}
						}
					}
				}
				for( ;; ) {
					i = picknumber(w);
					j = picknumber(w);
					if( buf[i*(w+1)+j] == ' ' && morethanonecandidates(i,j,c) )
						break;
					if( pulsar() < 0 )
						goto panic;
				}
	choose:
				if( symmetric )
					++k;
				if( solvedbuf == buf ) {
					x = picknumber(strlen(c));
					buf[i*(w+1)+j] = c[x];
				}
				else {
					buf[i*(w+1)+j] = solvedbuf[i*(w+1)+j];
				}
				++e;
			}
			else {
				/* unsolvable, try another number */
				do {
					for( ;; ) {
						i = picknumber(w);
						j = picknumber(w);
						if( buf[i*(w+1)+j] != ' ' )
							break;
						if( pulsar() < 0 )
							goto panic;
					}
					buf[i*(w+1)+j] = ' ';
					--e;
				} while( e * 2 > w * w );	/* to many cels, can never be a challenging sudoku */
				if( symmetric ) {
					if( sudokusize < 5 )
						--k;	/* avoid finding this via symmetric search */
					else
						k -= w;	/* for a long time generate randomly */
				}
			}
			init_sudoku(sudokusize, buf);
			z=1;
		}
	}
	/* buf now contains a solvable sudoku, let it hang out... */
	pulsing = 0;
	on_clear_activate(0, 0);
	for( i=0; i<w; i++ ) {
		y = i % sudokusize;
		z = i / sudokusize;
		for( j=0; j<w; j++ ) {
			if( buf[i*(w+1)+j] != ' ' ) {
				x = j % sudokusize;
				k = j / sudokusize;
				event.keyval = buf[i*(w+1)+j];
				gtk_widget_grab_focus(button[z][k][y][x]);
				event.type = GDK_KEY_PRESS;
				event.window = gtk_widget_get_root_window(windowGnuSudoku);
				event.send_event = TRUE;
				on_button_key_press_event(button[z][k][y][x],&event,0);
			}
		}
	}
	sprintf(c,"This is a grade %d sudoku",grade);
	gtk_label_set_text(GTK_LABEL(labelExplain),c);
panic:
	if( solvedbuf != buf )
		free(solvedbuf);
	free(buf);
}

void
on_buttonNew_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
	on_new1_activate(0,0);
}


void
on_buttonOpen_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	on_open1_activate(0,0);
}


void
on_buttonGenerate_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
	on_generate1_activate(0,0);
}


void
on_buttonSave_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	on_save1_activate(0,0);
}


void
on_buttonSaveAs_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
	on_save_as1_activate(0,0);
}


void
on_buttonSetup_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
	on_setup_activate(0,0);
}


void
on_buttonHint_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	on_hint_activate(0,0);
}


void
on_buttonSolve_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
	on_solve_activate(0,0);
}


void
on_buttonClear_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
	on_clear_activate(0,0);	/* misuse menu click routine */
}


void
on_buttonHelp_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
	on_help1_activate(0,0);
}

void
on_help1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
	char buf[160];
	sprintf(buf,"%s " PACKAGE_DATA_DIR "/" PACKAGE "/doc/index.html&", helpbrowser);
	system(buf);
}

