#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>

#include <libintl.h>
#define _(String) gettext (String)

#include "sudoku.h"

static int sudokurib = 3;
static int sudokuwidth = 9;
static char validchars[26];
int sudokuvalid = 1;
typedef char CANDIDATESROW[25][26];
CANDIDATESROW *candidates = 0;

char sudoku[25][25];

#define G_R_Y_X_CODED(g,r,y,x) 	G_R_Y_X_V_CODED(g,r,y,x,sudoku[y][x]);

/* forward declaration */
int travelsudoku(void);

int getcel(int i, int j)
{
	return sudoku[i][j];
}

void init_sudoku(int rib, const char *cels)
{
	int x,y,c;
	const char *p;
	if( rib < 2 || rib > 5 )
		return;
	sudokuvalid = 1;
	sudokurib = rib;
	sudokuwidth = rib * rib;
	switch(rib) {
		case 2:
			strcpy(validchars,"1234");
			break;
		case 3:
		default:
			strcpy(validchars,"123456789");
			break;
		case 4:
			strcpy(validchars,"0123456789ABCDEF");
			break;
		case 5:
			strcpy(validchars,"ABCDEFGHIJKLMNOPQRSTUVWXY");
			break;
	}
	if( candidates )
		free(candidates);
	candidates = malloc(sudokuwidth*sizeof(CANDIDATESROW));
	if( !cels )
		return;
	p = cels;
	for( y=0; y<sudokuwidth; y++ ) {
		while( *p && *p < ' ' )
			p++;
		for( x=0; x<sudokuwidth; x++ ) {
			c = toupper(*p);
			if( strchr(validchars,c) )
				sudoku[y][x] = c;
			else
				sudoku[y][x] = ' ';
			if( *p && *p >= ' ' )
				p++;
		}
	}
}

int unsolved(void)
{
	int i,j;
	for( i=0; i<sudokuwidth; i++ )
		for( j=0; j<sudokuwidth; j++ )
			if( sudoku[i][j]==' ' )
				return 1;
	return 0;
}

static void deletechar(char *candidates, char c)
{
	char *p = strchr(candidates,c);
	if( p )
		strcpy(p,p+1);
}

int getcandidates(int i, int j, char *candidates)
{
	int h,k,r,x,y,z;
	if( sudoku[i][j] == ' ' ) {
		strcpy(candidates,validchars);
		x = i/sudokurib;
		y = j/sudokurib;
		x *= sudokurib;
		y *= sudokurib;
		for( z=0; z<sudokuwidth; z++ ) {
			if( !(z == i || sudoku[z][j] == ' ') )
				deletechar(candidates, sudoku[z][j]);
			if( !(z == j || sudoku[i][z] == ' ') )
				deletechar(candidates, sudoku[i][z]);
		}
		for( z=0; z<sudokurib; z++ ) {
			k = x + z;
			for( h=0; h<sudokurib; h++ ) {
				r = y + h;
				if( !((k ==  i && r == j) || sudoku[k][r] == ' ') )
					deletechar(candidates, sudoku[k][r]);
			}
		}
		k = strlen(candidates);
		if( k==1 )
			sudoku[i][j] = candidates[0];
		if( k==0 )
			sudokuvalid = 0;
		return k;
	}
	sprintf(candidates,"%c",sudoku[i][j]);
	return -1;
}

int morethanonecandidates(int i, int j, char *candidates)
{
	int h,k,r,x,y,z;
	if( sudoku[i][j] == ' ' ) {
		strcpy(candidates,validchars);
		x = i/sudokurib;
		y = j/sudokurib;
		x *= sudokurib;
		y *= sudokurib;
		for( z=0; z<sudokuwidth; z++ ) {
			if( !(z == i || sudoku[z][j] == ' ') )
				deletechar(candidates, sudoku[z][j]);
			if( !(z == j || sudoku[i][z] == ' ') )
				deletechar(candidates, sudoku[i][z]);
		}
		for( z=0; z<sudokurib; z++ ) {
			k = x + z;
			for( h=0; h<sudokurib; h++ ) {
				r = y + h;
				if( !((k ==  i && r == j) || sudoku[k][r] == ' ') )
					deletechar(candidates, sudoku[k][r]);
			}
		}
		return strlen(candidates) > 1;
	}
	return 0;
}

int excludecandidates(int r,int k,char *c, char *excl)
{
	int i;
	for( i=0; excl[i]; i++ )
		deletechar(c, excl[i]);
	i = strlen(c);
	if( i==1 )
		sudoku[r][k] = c[0];
	return i;
}

typedef enum {
	NOREASON,
	SINGLECANDIDATE,
	NOWHEREELSEINCOL,
	NOWHEREELSEINROW,
	NOWHEREELSEINSQUARE,
	TRIOPAIREXCLUDEDINCOL,
	TRIOPAIREXCLUDEDINROW,
	TRIOPAIREXCLUDEDINSQUARE,
	JUSTBECAUSE
	} REASON;
	
const char *explanation(int hint)
{
	REASON r = R_DECODED(hint);
	switch( r ) {
		case NOREASON:
			return _("none");
		case SINGLECANDIDATE:
			return _("no other possibility");
		case NOWHEREELSEINCOL:
			return _("only possible place in column");
		case NOWHEREELSEINROW:
			return _("only possible place in row");
		case NOWHEREELSEINSQUARE:
			return _("only possible place in square");
		case TRIOPAIREXCLUDEDINCOL:
			return _("the other two possibilities occur two times elsewhere in the column");
		case TRIOPAIREXCLUDEDINROW:
			return _("the other two possibilities occur two times elsewhere in the row");
		case TRIOPAIREXCLUDEDINSQUARE:
			return _("the other two possibilities occur two times elsewhere in the square");
		case JUSTBECAUSE:
			return _(" ");
		default:
			break;
	}
	return _("i say!!!");
}

int trytosolve(int i, int j)
{
	char c[3][25],*cp[3][25];
	unsigned char cx[3][25],cy[3][25];
	int h,k,r,t,x,y,z;	/* z: (indexes) 0 for rows, 1 for columns and 2 for squares */
	/* build row col and square administration */
	x = i/sudokurib;
	y = j/sudokurib;
	x *= sudokurib;
	y *= sudokurib;
	h = 0;
	for( r=0; r<sudokurib; r++ )
		for( k=0; k<sudokurib; k++ ) {
			cx[0][h] = h;
			cy[0][h] = j;
			cx[1][h] = i;
			cy[1][h] = h;
			cx[2][h] = x+r;
			cy[2][h] = y+k;
			for( z=0; z<3; z++ ) {
				cp[z][h] = candidates[cx[z][h]][cy[z][h]];
				c[z][h] = strlen(cp[z][h]);
			}
			h++;
		}
	/* look for candidates with numbers that don't occur anywhere else */
	for( z=0; z<3; z++ ) {		/* search rows, cols, squares */
		for( r=0; r<sudokuwidth; r++ ) {
			if( c[z][r] > 1 ) {
				for( h=0; cp[z][r][h]; h++ ) {
					t=0;
					for( k=0; k<sudokuwidth && t==0; k++ ) {
						if( k==r )
							continue;
						if( strchr(cp[z][k],cp[z][r][h]) )
							t++;
					}
					if( t==0 ) {
						x = cx[z][r];
						y = cy[z][r];
						sudoku[x][y] = cp[z][r][h];
						return G_R_Y_X_CODED(2,NOWHEREELSEINCOL+z,x,y);
					}
				}
			}
		}
	}
	/* look for 2 pair with equal candidates */
	for( z=0; z<3; z++ ) {		/* search rows, cols, squares */
		for( r=0; r<sudokuwidth; r++ ) {
			if( c[z][r] == 2 ) {
				for( k=r+1; k<sudokuwidth; k++ ) {
					if( c[z][k] == 2 && strcmp(cp[z][r],cp[z][k]) == 0 ) {
						for( h=0; h<sudokuwidth; h++ ) {
							if( h==r || h==k )
								continue;
							if( c[z][h] > 1 && excludecandidates(cx[z][h],cy[z][h],cp[z][h],cp[z][k])==1 )
								return G_R_Y_X_CODED(3,TRIOPAIREXCLUDEDINCOL+z,cx[z][h],cy[z][h]);
						}
					}
				}
			}
		}
	}
	return 0;
}

int travelsudoku(void)
{
	int i,j,t;
	if( !candidates )
		return 0;
	for( i=0; i<sudokuwidth; i++ )
		for( j=0; j<sudokuwidth; j++ ) {
			t = getcandidates(i,j,candidates[i][j]);
			if( t==1 )
				return G_R_Y_X_CODED(1,SINGLECANDIDATE,i,j);
			if( !sudokuvalid )
				return 0;
		}
	for( i=0; i<sudokuwidth; i++ )
		for( j=0; j<sudokuwidth; j++ )
			if( sudoku[i][j] == ' ' )
				if( (t=trytosolve(i,j)) )
					return t;
	return 0;
}

/*
 * build up a solved sudoku
 * 3x3 example:
 *
 *   123 456 789
 *   456 789 123
 *   789 123 456
 *
 *   234 567 891
 *   567 891 234
 *   891 234 567
 *
 *   345 678 912
 *   678 912 345
 *   912 345 678
 */
static void initialsudoku(void)
{
	int r,k,i;
	for( r=0; r<sudokuwidth; r++ ) {
		i = (r*sudokurib)%sudokuwidth;
		i += r/sudokurib;
		for( k=0; k<sudokuwidth; k++ ) {
			sudoku[r][k] = validchars[i];
			i = (i+1)%sudokuwidth;
		}
	}
}

static void shuffleincolumns(char v1, char v2)
{
	int r1=0,r2=0;
	int i,j,t;
	for( i=0; i<sudokuwidth; i++ ) {
		t = 0;
		for( j=0; t<2 && j<sudokuwidth; j++ ) {
			if( sudoku[j][i] == v1 ) {
				r1 = j;
				t++;
			}
			if( sudoku[j][i] == v2 ) {
				r2 = j;
				t++;
			}
		}
		sudoku[r1][i] = v2;
		sudoku[r2][i] = v1;
	}
}

static void shuffleinrows(char v1, char v2)
{
	int r1=0,r2=0;
	int i,j,t;
	for( i=0; i<sudokuwidth; i++ ) {
		t = 0;
		for( j=0; t<2 && j<sudokuwidth; j++ ) {
			if( sudoku[i][j] == v1 ) {
				r1 = j;
				t++;
			}
			if( sudoku[i][j] == v2 ) {
				r2 = j;
				t++;
			}
		}
		sudoku[i][r1] = v2;
		sudoku[i][r2] = v1;
	}
}


static void shuffleinsquares(char v1, char v2)
{
	int r1=0,r2=0,k1=0,k2=0;
	int h,i,j,r,k,t;
	for( i=0; i<sudokuwidth; i++ ) {
		j = i/sudokurib;
		j *= sudokurib;
		h = i % sudokurib;
		h *= sudokurib;
		t = 0;
		for( r=0; t<2 && r<sudokurib; r++ )
			for( k=0; t<2 && k<sudokurib; k++ ) {
				if( sudoku[r+j][k+h] == v1 ) {
					r1 = r+j;
					k1 = k+h;
					t++;
				}
				if( sudoku[r+j][k+h] == v2 ) {
					r2 = r+j;
					k2 = k+h;
					t++;
				}
			}
		sudoku[r1][k1] = v2;
		sudoku[r2][k2] = v1;
	}
}

static void switchcolumns(int k1, int k2)
{
	int i,t;
	for( i=0; i<sudokuwidth; i++ ) {
		t = sudoku[i][k1];
		sudoku[i][k1] = sudoku[i][k2];
		sudoku[i][k2] = t;
	}
}

static void switchrows(int r1, int r2)
{
	int i,t;
	for( i=0; i<sudokuwidth; i++ ) {
		t = sudoku[r1][i];
		sudoku[r1][i] = sudoku[r2][i];
		sudoku[r2][i] = t;
	}
}

int picknumber(int high)
{
	int r,res;
	r = random();
	res = (r>>10) % high;
	if( res < 0 || res >= high ) {
		printf("bug in c runtime!! (%d>>10) %% %d results in %d\n",
			r, high, res);
		return 0;
	}
	return res; 
}

static void generatesolved5x5(int (*cb)(void))
{
	int i,j,k,t;
	initialsudoku();
	t = picknumber(sudokuwidth<<4)*10;
	while( t-- > 0 ) {
		if( cb && (cb() < 0) )
			return;
		k = picknumber(5);
		i = 0;
		j = 0;
		while( i==j ) {
			i = picknumber(sudokuwidth);
			j = picknumber(sudokuwidth);
		}
		switch(k) {
			case 0:
				shuffleincolumns(validchars[i], validchars[j]);
				break;
			case 1:
				shuffleinrows(validchars[i], validchars[j]);
				break;
			case 2:
				shuffleinsquares(validchars[i], validchars[j]);
				break;
			case 3:
				j = (i/sudokurib)*sudokurib + j%sudokurib;
				switchcolumns(i, j);
				break;
			case 4:
			default:
				j = (i/sudokurib)*sudokurib + j%sudokurib;
				switchrows(i, j);
				break;
		}
	}
}

static int validcharpos(int x)
{
	char *p;
	p = strchr(validchars,x);
	if( p )
		return p-validchars;
	return 0;
}

void generatesolved(int (*cb)(void))
{
	int h,i,j,k,r,x,y,z,t,bad,tries;
	char a[25][25][26],n[25],pool[25];
	z = sudokuwidth*sudokuwidth;
	tries = 0;
	do {
		if( tries++ > 10 && sudokurib > 4 ) {
			generatesolved5x5(cb);
			return;
		}
		bad = 0;
		for( i=0; i<sudokuwidth; i++ ) {
			for( j=0; j<sudokuwidth; j++ ) {
				sudoku[i][j] = ' ';
				strcpy(a[i][j],validchars);
			}
		}
		if( cb && (cb() < 0) )
			return;
		for( t=0; t<z && !bad; t++ ) {
			/* first find the cel with the least candidates */
			h = 26;
			r = 0;
			k = 0;
			memset(n,0,sudokuwidth);
			for( i=0; i<sudokuwidth && h>0; i++ ) {
				for( j=0; j<sudokuwidth && h>0; j++ ) {
					x=sudoku[i][j];
					if( x == ' ' ) {
						if( h > strlen(a[i][j]) ) {
							h = strlen(a[i][j]);
							r = i;
							k = j;
						}
					}
					else {
						n[validcharpos(x)]++;
					}
				}
			}
			if( cb && (cb() < 0) )
				return;
			if( h < 2 ) {
				if( h )
					sudoku[r][k] = a[r][k][0];
				else {
					sudoku[r][k] = validchars[0];
					bad++;
				}
			}
			else {
				if( t < sudokuwidth * sudokurib ) {
					j = picknumber(h);
					sudoku[r][k] = a[r][k][j];
				}
				else {
					/* pick the character that occurs the least */
					y = 26;
					j = 0;
					x = 0;
					for( i=0; i<sudokuwidth; i++ ) {
						if( strchr(a[r][k],validchars[i]) ) {
							if( y >= n[i] ) {
								if( y > n[i] )
									x = 0;
								y = n[i];
								j = i;
								pool[x] = validchars[i];
								x++;
							}
						}
					}
					j = picknumber(x);
					sudoku[r][k] = pool[j];
				}
			}
			/* update candidates */
			h = sudoku[r][k];
			/* row and col and square */
			y = r / sudokurib;
			x = k / sudokurib;
			y *= sudokurib;
			x *= sudokurib;
			for( i=0; i<sudokuwidth; i++ ) {
				deletechar(a[i][k],h);
				deletechar(a[r][i],h);
				deletechar(a[y+(i/sudokurib)][x+(i%sudokurib)],h);
			}
			if( cb && (cb() < 0) )
				return;
		}
	} while(bad);
}

char *sudokulines(void)		/* return one long string with rows seperated by newline */
{
	char *buf,*p;
	int i,j,z;
	z = sudokuwidth*sudokuwidth+sudokuwidth+1;
	buf = malloc(z);
	if( !buf )
		return 0;
	p = buf;
	for( i=0; i<sudokuwidth; i++ ) {
		for( j=0; j<sudokuwidth; j++ ) {
			z = getcel(i,j);
			if( ispunct(z) )
				*p++ = ' ';
			else
				*p++ = toupper(z);
		}
		*p++ = '\n';
	}
	*p = '\0';
	return buf;
}

#ifdef TEST_SUDOKU

void printsudoku(const char *head, int n, int c, int y, int x, const char *r)
{
	int i,j;
	printf(head,n,c,y,x,r);
	for( i=0; i<sudokuwidth; i++ ) {
		if( (i%sudokurib) == 0 ) {
			for( j=0; j<sudokuwidth; j++ ) {
				if( (j%sudokurib) == 0 )
					putchar('+');
				putchar('-');
			}
			printf("+\n");
		}
		for( j=0; j<sudokuwidth; j++ ) {
			if( (j%sudokurib) == 0 )
				putchar('|');
			putchar(sudoku[i][j]);
		}
		printf("|\n");
	}
	for( j=0; j<sudokuwidth; j++ ) {
		if( (j%sudokurib) == 0 )
			putchar('+');
		putchar('-');
	}
	printf("+\n");
}

int main(int argc, const char **argv)
{
	int i,j,k;
	char line[80];
	srandom(time(0));
	i = 3;
	k = 0;
	if( argc==2 && argv[1][0]=='-' ) {
		switch(argv[1][1]) {
			case '0':
				i = 2;
				k = 1;
				break;
			case '1':
				i = 4;
				k = 1;
				break;
			case '2':
			case '3':
			case '4':
			case '5':
				i = argv[1][1]-'0';
				break;
			case '!':
				i = 5;
				k = 1;
				break;
			default:
				i = 3;
				k = 1;
				break;
		}
	}
	init_sudoku(i,0);
	if( k ) {
		initialsudoku();
		printsudoku("Situation 0 -- Initial\n",0,0,0,0,0);
		generatesolved(0);
	}
	else {
		for( i=0; i<sudokuwidth; i++ ) {
			if( isatty(STDIN_FILENO) )
				printf("Enter row %d:",i+1);
			strcpy(line,".....zzzzz.....zzzzz.....zzzzzz");
			if( !fgets(line,80,stdin) )
				return 1;
			for( j=0; j<sudokuwidth; j++ ) {
				if( strchr(validchars,toupper(line[j])) && line[j] )
					sudoku[i][j] = toupper(line[j]);
				else 
					sudoku[i][j] = ' ';
			}
		}
	}
	k=1;
	j=G_R_Y_X_V_CODED(0,JUSTBECAUSE,0,0,' ');
	while( unsolved() && sudokuvalid ) {
		if( R_DECODED(j) == JUSTBECAUSE )
			printsudoku("Situation %4d -- Unsolved\n",k,0,0,0,0);
		else
			printsudoku("Situation %4d -- Unsolved, %c at row %d col %d, %s\n",
				k, V_DECODED(j), Y_DECODED(j)+1, X_DECODED(j)+1, explanation(j));
		++k;
		j = travelsudoku();
	}
	if( sudokuvalid )
		printsudoku("Situation %4d -- Solved\n",k,0,0,0,0);
	else
		printsudoku("Situation %4d -- unsolveable\n",k,0,0,0,0);
	free(candidates);
	return 0;
}

#endif
